import os
import sys
import types

# Ensure package source is importable
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../src")))

# Stub macOS apple watcher extension so ecotorch.watcher can import on Darwin
APPLE_MODULE = "ecotorch.watcher.apple"
if APPLE_MODULE not in sys.modules:
    apple_mod = types.ModuleType(APPLE_MODULE)

    class Monitor:
        def __init__(self, *args, **kwargs):
            pass

        def get_current_power(self):
            return 0.0

        def get_gpu_utilization(self):
            return "0%"

    apple_mod.Monitor = Monitor
    sys.modules[APPLE_MODULE] = apple_mod

# Stub external runtime-only deps used in geolocator to avoid import errors in tests
if "geoip2fast" not in sys.modules:
    geoip_mod = types.ModuleType("geoip2fast")

    class GeoIP2Fast:  # minimal stub, tests will patch methods as needed
        def lookup(self, ip):
            class _R:
                country_name = "World"

            return _R()

    geoip_mod.GeoIP2Fast = GeoIP2Fast
    sys.modules["geoip2fast"] = geoip_mod

if "pycountry" not in sys.modules:
    pc_mod = types.ModuleType("pycountry")

    class _Countries:
        def search_fuzzy(self, name):
            return []

    pc_mod.countries = _Countries()
    sys.modules["pycountry"] = pc_mod

# Stub pynvml to allow importing Linux/Windows monitors if needed
if "pynvml" not in sys.modules:
    pynvml_mod = types.ModuleType("pynvml")

    class NVMLError(Exception):
        pass

    def nvmlInit():
        return None

    def nvmlDeviceGetCount():
        return 0

    def nvmlDeviceGetHandleByIndex(i):
        return None

    def nvmlDeviceGetPowerUsage(handle):
        return 0

    class _Util:
        gpu = 0

    def nvmlDeviceGetUtilizationRates(handle):
        return _Util()

    pynvml_mod.NVMLError = NVMLError
    pynvml_mod.nvmlInit = nvmlInit
    pynvml_mod.nvmlDeviceGetCount = nvmlDeviceGetCount
    pynvml_mod.nvmlDeviceGetHandleByIndex = nvmlDeviceGetHandleByIndex
    pynvml_mod.nvmlDeviceGetPowerUsage = nvmlDeviceGetPowerUsage
    pynvml_mod.nvmlDeviceGetUtilizationRates = nvmlDeviceGetUtilizationRates

    sys.modules["pynvml"] = pynvml_mod



import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import datasets, transforms

from ecotorch import TrainTracker


# 1. Define a simple CNN for MNIST
class SimpleCNN(nn.Module):
    def __init__(self):
        super(SimpleCNN, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = torch.relu(torch.max_pool2d(self.conv1(x), 2))
        x = torch.relu(torch.max_pool2d(self.conv2(x), 2))
        x = x.view(-1, 320)
        x = torch.relu(self.fc1(x))
        x = self.fc2(x)
        return torch.log_softmax(x, dim=1)


def main():
    # 2. Setup device and data
    device = "cuda" if torch.cuda.is_available() else "cpu"
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )

    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST("./data", train=True, download=True, transform=transform),
        batch_size=64,
        shuffle=True,
    )

    model = SimpleCNN().to(device)
    optimizer = optim.SGD(model.parameters(), lr=0.01)
    criterion = nn.CrossEntropyLoss()
    epochs = 1

    print("Starting tracked training...")

    # 3. Use TrainTracker as a context manager
    with TrainTracker(
        model=model, epochs=epochs, train_dataloader=train_loader
    ) as tracker:
        initial_loss = None

        for epoch in range(epochs):
            model.train()
            for batch_idx, (data, target) in enumerate(train_loader):
                data, target = data.to(device), target.to(device)
                optimizer.zero_grad()
                output = model(data)
                loss = criterion(output, target)
                loss.backward()
                optimizer.step()

                if initial_loss is None:
                    initial_loss = loss.item()

                if batch_idx % 100 == 0:
                    print(
                        f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)}] Loss: {loss.item():.6f}"
                    )

        final_loss = loss.item()

    # 4. Calculate and display efficiency score
    score = tracker.calculate_efficiency_score(
        initial_loss=initial_loss, final_loss=final_loss
    )
    print("\n--- Training Results ---")
    print(f"Total Energy Used: {tracker.used_energy} kWh")
    print(f"Total Time: {tracker.total_time} seconds")
    print(f"Efficiency Score: {score}")


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from pprint import pprint

import pandas as pd
import pycountry

data = {}
price_units = {}
conversion = {
    "EUR": 1.17955,
    "INR": 0.010992,
    "AUD": 0.705465,
    "GBP": 1.348895,
    "BGN": 0.60,
    "WON": 0.00069,
    "JPY": 0.0065,
    "CAD": 0.73,
    "BRL": 0.19,
    "SGD": 0.79,
    "RON": 0.23,
    "USD": 1,
}


def convert_country_to_iso(country_name: str):
    try:
        result = pycountry.countries.search_fuzzy(country_name)[0]
        return result.alpha_3
    except (LookupError, IndexError):
        return "World"


info_table = pd.read_csv("./src/ecotorch/wallet/infotable.csv")
p = Path("./src/ecotorch/wallet/datasets")
for file_path in p.glob("*.csv"):
    df = pd.read_csv(file_path)
    name = str(file_path.name).split("_")[0]
    mask = info_table["Country"] == name
    try:
        unit = str(info_table.loc[mask, "Price Unit"].iloc[0])
        price_units[name] = unit.split("/")
    except IndexError:
        print(name)

    try:
        value = float(df[name].tail(1).item())
    except KeyError:
        value = float(df.iloc[-1, 1])

    if price_units[name][1] == "MWh":
        value /= 1000

    value *= conversion[price_units[name][0]]

    name = convert_country_to_iso(name)

    data[name] = value


data["World"] = 0.162
base_path = Path(__file__).parent
file_path = base_path / "electricity_price.json"

# Open the file in write mode and dump the data
with open(file_path, "w") as json_file:
    json.dump(data, json_file, indent=4)



from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from ecotorch.datahandler import DataHandler


@pytest.fixture
def mock_df():
    data = {
        "Area": ["United States", "United States", "Austria"],
        "ISO 3 code": ["USA", "USA", "AUT"],
        "Variable": ["CO2 intensity", "CO2 intensity", "CO2 intensity"],
        "Year": [2020, 2021, 2021],
        "Value": [0.4, 0.35, 0.2],
    }
    return pd.DataFrame(data)


@pytest.fixture
def data_handler(mock_df):
    # Ensure a deterministic initial location stored in the handler
    with (
        patch("pandas.read_csv", return_value=mock_df),
        patch("ecotorch.datahandler.get_location", return_value="USA"),
    ):
        handler = DataHandler()
        return handler


def test_get_intensity_mocked_location(data_handler):
    # Uses the stored location from initialization ('USA')
    assert data_handler.get_intensity() == 0.35


def test_get_intensity_specific_country(data_handler):
    with patch("ecotorch.datahandler.convert_country_to_iso", return_value="USA"):
        assert data_handler.get_intensity("United States") == 0.35


def test_get_intensity_invalid_country_code(data_handler):
    with patch("ecotorch.datahandler.convert_country_to_iso", return_value=None):
        with pytest.raises(ValueError, match="Cannot find country code"):
            data_handler.get_intensity("NonExistent")


def test_get_intensity_no_data(data_handler):
    with patch("ecotorch.datahandler.convert_country_to_iso", return_value="ZZZ"):
        # ZZZ is not in mock_df
        with pytest.raises(ValueError, match="No data for the country"):
            data_handler.get_intensity("Unknown Country")


def test_get_countries(data_handler):
    assert data_handler.countries == ["Austria", "United States"]


def test_get_intensity_world_success():
    # World path should use latest available year
    data = {
        "Area": ["World", "World"],
        "ISO 3 code": [None, None],
        "Variable": ["CO2 intensity", "CO2 intensity"],
        "Year": [2020, 2021],
        "Value": [0.5, 0.4],
    }
    df = pd.DataFrame(data)
    with (
        patch("pandas.read_csv", return_value=df),
        patch("ecotorch.datahandler.get_location", return_value="World"),
    ):
        handler = DataHandler()
        assert handler.get_intensity() == 0.4


def test_get_intensity_world_missing_data_raises():
    data = {
        "Area": ["United States"],
        "ISO 3 code": ["USA"],
        "Variable": ["CO2 intensity"],
        "Year": [2021],
        "Value": [0.35],
    }
    df = pd.DataFrame(data)
    with (
        patch("pandas.read_csv", return_value=df),
        patch("ecotorch.datahandler.get_location", return_value="World"),
    ):
        handler = DataHandler()
        with pytest.raises(ValueError, match="Couldn't find values for the world"):
            handler.get_intensity()


def test_get_latest_year_by_country_code(data_handler):
    assert data_handler._get_latest_year_by_country_code("USA") == 2021
    assert data_handler._get_latest_year_by_country_code("AUT") == 2021
    assert data_handler._get_latest_year_by_country_code("ZZZ") is None



import warnings

import pycountry
import requests
from geoip2fast import GeoIP2Fast


def get_ip() -> str:
    try:
        public_ip = requests.get("https://api.ipify.org").text
    except requests.RequestException:
        raise requests.ConnectionError(
            "Cannot find public IP, manually determine a country."
        )
    return public_ip


def get_location() -> str:
    warnings.filterwarnings("ignore")
    geoip = GeoIP2Fast()
    try:
        found_name = geoip.lookup(get_ip()).country_name
        return convert_country_to_iso(found_name)
    except requests.RequestException:
        return "World"


def convert_country_to_iso(country_name: str):
    try:
        result = pycountry.countries.search_fuzzy(country_name)[0]
        return result.alpha_3
    except (LookupError, IndexError):
        return "World"



import torch
import torch.nn as nn
from torchvision import datasets, transforms

from ecotorch import EvalTracker


# 1. Define a simple CNN for MNIST (same as in basic_training.py)
class SimpleCNN(nn.Module):
    def __init__(self):
        super(SimpleCNN, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = torch.relu(torch.max_pool2d(self.conv1(x), 2))
        x = torch.relu(torch.max_pool2d(self.conv2(x), 2))
        x = x.view(-1, 320)
        x = torch.relu(self.fc1(x))
        x = self.fc2(x)
        return torch.log_softmax(x, dim=1)


def main():
    # 2. Setup device and data
    device = "cuda" if torch.cuda.is_available() else "cpu"
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )

    test_loader = torch.utils.data.DataLoader(
        datasets.MNIST("./data", train=False, transform=transform),
        batch_size=1000,
        shuffle=True,
    )

    # Assume a pre-trained model
    model = SimpleCNN().to(device)
    model.eval()

    print("Starting tracked evaluation...")

    # 3. Use EvalTracker as a context manager
    with EvalTracker(test_dataloader=test_loader, model=model) as tracker:
        correct = 0
        total = 0
        with torch.no_grad():
            for data, target in test_loader:
                data, target = data.to(device), target.to(device)
                output = model(data)
                pred = output.argmax(dim=1, keepdim=True)
                correct += pred.eq(target.view_as(pred)).sum().item()
                total += target.size(0)

        accuracy = correct / total

        # 4. Calculate and display efficiency score
        score = tracker.calculate_efficiency_score(accuracy=accuracy)
        print("\n--- Evaluation Results ---")
        print(f"Accuracy: {accuracy * 100:.2f}%")
        print(f"Total Energy Used: {tracker.used_energy} kWh")
        print(f"Total Time: {tracker.total_time} seconds")
        print(f"Efficiency Score: {score}")


if __name__ == "__main__":
    main()



"""
A lightweight package to measure the ecological and financial effect of training and evaluation of pytorch projects.
"""

from ._geolocator import get_ip, get_location
from .core import EvalTracker, Mode, TrainTracker, evaluate, train
from .datahandler import DataHandler

__version__ = "0.2.6"
__author__ = "David Leonard Nagy"
__copyright__ = "Copyright 2026, David Leonard Nagy"
__credits__ = ["David Leonard Nagy"]
__license__ = "GNU LGPLv3"
__maintainer__ = "David Leonard Nagy"
__email__ = "nagy.david.leonard@gmail.com"
__status__ = "Development"

__all__ = ["TrainTracker", "EvalTracker", "DataHandler", "evaluate", "train", "Mode"]



from unittest.mock import MagicMock, patch

import pytest
import requests

import ecotorch._geolocator as geolocator


def test_get_ip_success():
    with patch("requests.get") as mock_get:
        mock_get.return_value.text = "1.2.3.4"
        assert geolocator.get_ip() == "1.2.3.4"


def test_get_ip_failure():
    with patch("requests.get", side_effect=requests.RequestException):
        with pytest.raises(requests.ConnectionError):
            geolocator.get_ip()


def test_convert_country_to_iso_success():
    with patch("ecotorch._geolocator.pycountry.countries.search_fuzzy") as mock_search:
        mock_country = MagicMock()
        mock_country.alpha_3 = "USA"
        mock_search.return_value = [mock_country]

        assert geolocator.convert_country_to_iso("United States") == "USA"


def test_convert_country_to_iso_failure():
    with patch(
        "ecotorch._geolocator.pycountry.countries.search_fuzzy", side_effect=LookupError
    ):
        assert geolocator.convert_country_to_iso("Unknown Country") == "World"


def test_get_location_success():
    with (
        patch("ecotorch._geolocator.get_ip", return_value="1.2.3.4"),
        patch("ecotorch._geolocator.GeoIP2Fast") as mock_geoip_cls,
        patch(
            "ecotorch._geolocator.convert_country_to_iso", return_value="USA"
        ) as mock_convert,
    ):

        mock_geoip_instance = mock_geoip_cls.return_value
        mock_geoip_instance.lookup.return_value.country_name = "United States"

        assert geolocator.get_location() == "USA"
        mock_convert.assert_called_with("United States")


def test_get_location_requests_failure():
    # If underlying IP or lookup fails with requests.RequestException, we fall back to 'World'
    with patch("ecotorch._geolocator.get_ip", side_effect=requests.RequestException):
        assert geolocator.get_location() == "World"



import json
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest

from ecotorch.wallet.extractor import Extractor


def test_extractor_init():
    with patch("ecotorch.wallet.extractor.Path") as mock_path:
        # Mocking the path to electricity_price.json
        mock_file_path = mock_path.return_value / "electricity_price.json"

        # Mocking json load
        mock_data = {"USA": 0.12, "DEU": 0.35}
        with patch(
            "ecotorch.wallet.extractor.open", mock_open(read_data=json.dumps(mock_data))
        ):
            extractor = Extractor()
            assert extractor.prices == mock_data


def test_extractor_calculate_cost():
    mock_data = {"USA": 0.12, "DEU": 0.35}
    with patch(
        "ecotorch.wallet.extractor.Extractor._load_json", return_value=mock_data
    ):
        extractor = Extractor()

        # USA: 0.12 * 10 = 1.2
        assert extractor.calculate_cost("USA", 10) == 1.2
        # DEU: 0.35 * 2 = 0.7
        assert extractor.calculate_cost("DEU", 2) == 0.7
        # Non-existent country should raise KeyError (based on implementation)
        with pytest.raises(KeyError):
            extractor.calculate_cost("FRA", 1)


def test_extractor_load_json_errors():
    with patch("ecotorch.wallet.extractor.open", side_effect=FileNotFoundError):
        extractor = Extractor.__new__(Extractor)
        with pytest.raises(
            FileNotFoundError, match="Internal error, please contact the developers."
        ):
            extractor._load_json("fake_path")

    with patch("ecotorch.wallet.extractor.open", mock_open(read_data="invalid json")):
        extractor = Extractor.__new__(Extractor)
        with pytest.raises(
            json.JSONDecodeError, match="Internal error, please contact developers."
        ):
            extractor._load_json("fake_path")



from unittest.mock import patch

import pytest
import torch
import torch.nn as nn
import torch.optim as optim

from ecotorch.core import (
    EvalTracker,
    TrainTracker,
    _transfer_optimizer_to_device,
    evaluate,
    train,
)


class SimpleModel(nn.Module):
    def __init__(self):
        super(SimpleModel, self).__init__()
        self.fc = nn.Linear(10, 2)

    def forward(self, x):
        return self.fc(x)


@pytest.fixture
def mock_model():
    return SimpleModel()


@pytest.fixture
def mock_loader():
    # Create simple data
    inputs = torch.randn(5, 10)
    labels = torch.randint(0, 2, (5,))
    dataset = torch.utils.data.TensorDataset(inputs, labels)
    loader = torch.utils.data.DataLoader(dataset, batch_size=2)
    return loader


def test_transfer_optimizer_to_device():
    # Create a simple optimizer
    model = torch.nn.Linear(1, 1)
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

    # Mock optimizer state
    p = list(model.parameters())[0]
    optimizer.state[p] = {"step": 0, "momentum_buffer": torch.zeros_like(p)}

    # Just ensure no exceptions for CPU
    _transfer_optimizer_to_device(optimizer, "cpu")


def test_evaluate(mock_model, mock_loader):
    device = "cpu"
    accuracy = evaluate(mock_model, mock_loader, device)
    assert isinstance(accuracy, float)
    assert 0.0 <= accuracy <= 1.0


def test_train(mock_model, mock_loader):
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(mock_model.parameters(), lr=0.01)
    device = "cpu"

    # Train for 1 epoch
    trained_model, loss, first_loss, crit, opt = train(
        mock_model, criterion, optimizer, mock_loader, epoch=1, device=device
    )

    assert isinstance(trained_model, nn.Module)
    assert isinstance(loss, float)
    assert isinstance(first_loss, float)
    assert crit == criterion
    assert opt == optimizer


def test_traintracker_context_and_score(mock_model, mock_loader):
    with (
        patch("ecotorch.core._Monitor") as MockMonitor,
        patch("ecotorch.datahandler.DataHandler.get_intensity", return_value=0.5),
    ):

        with TrainTracker(
            model=mock_model, epochs=1, train_dataloader=mock_loader, country="USA"
        ) as tracker:
            assert tracker._start_time != 0
            MockMonitor.return_value.start.assert_called_once()
            # Simulate power samples in Watts collected each second
            tracker.energy_usage = [100.0, 120.0]

        MockMonitor.return_value.join.assert_called_once()
        assert tracker._is_closed
        # Score in [0,1]
        score = tracker.calculate_efficiency_score(initial_loss=2.0, final_loss=1.0)
        assert isinstance(score, float)
        assert 0.0 <= score <= 1.0
        assert tracker.total_cost >= 0


def test_evaltracker_context_and_score(mock_model, mock_loader):
    with (
        patch("ecotorch.core._Monitor") as MockMonitor,
        patch("ecotorch.datahandler.DataHandler.get_intensity", return_value=0.5),
    ):

        # Reuse TrainTracker's model/handler via constructor
        train_tracker = TrainTracker(
            model=mock_model, epochs=1, train_dataloader=mock_loader, country="USA"
        )
        with EvalTracker(
            test_dataloader=mock_loader, train_tracker=train_tracker
        ) as et:
            MockMonitor.return_value.start.assert_called_once()
            et.energy_usage = [50.0]
        MockMonitor.return_value.join.assert_called_once()

        score = et.calculate_efficiency_score(accuracy=0.85)
        assert isinstance(score, float)
        assert 0.0 <= score <= 1.0
        assert et.country == "USA"
        assert et.total_cost >= 0


def test_validation_errors(mock_model, mock_loader):
    with (
        patch("ecotorch.core._Monitor"),
        patch("ecotorch.datahandler.DataHandler.get_intensity", return_value=0.5),
    ):

        # Calling before context exit should raise
        tt = TrainTracker(model=mock_model, epochs=1, train_dataloader=mock_loader)
        with pytest.raises(RuntimeError):
            tt.calculate_efficiency_score(initial_loss=1.0, final_loss=0.5)

        # After exit, invalid accuracy bounds should raise
        with TrainTracker(
            model=mock_model, epochs=1, train_dataloader=mock_loader
        ) as tt2:
            tt2.energy_usage = [10.0]
        with pytest.raises(ValueError):
            tt2.calculate_efficiency_score(
                initial_loss=1.0, final_loss=0.5, accuracy=1.5
            )

        # EvalTracker: accuracy is required and must be in [0,1]
        with EvalTracker(test_dataloader=mock_loader, model=mock_model) as et:
            et.energy_usage = [10.0]
        with pytest.raises(ValueError):
            et.calculate_efficiency_score(accuracy=None)
        with pytest.raises(ValueError):
            et.calculate_efficiency_score(accuracy=-0.1)



import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset

from ecotorch import EvalTracker, TrainTracker

# Simple synthetic model and data
model = nn.Linear(10, 1)
data = torch.randn(100, 10)
target = torch.randn(100, 1)
dataset = TensorDataset(data, target)
loader = DataLoader(dataset, batch_size=10)


def advanced_demo():
    print("--- Advanced Usage Demo ---")

    # 1. Manually specifying a country (ISO 3-letter code)
    print("\n1. Tracking with a specific country (e.g., France - FRA)")
    with TrainTracker(
        model=model, epochs=1, train_dataloader=loader, country="FRA"
    ) as tracker:
        # Simulate training
        pass
    print(f"Detected/Specified Country: {tracker.country}")

    # 2. Reusing a TrainTracker for Evaluation
    print("\n2. Reusing TrainTracker for Evaluation")
    with TrainTracker(model=model, epochs=1, train_dataloader=loader) as train_tracker:
        # Simulate training
        pass

    # EvalTracker can take a train_tracker to inherit the model and data handler
    with EvalTracker(
        test_dataloader=loader, train_tracker=train_tracker
    ) as eval_tracker:
        # Simulate evaluation
        pass

    # 3. Accessing detailed metrics
    print("\n3. Accessing raw metrics")
    print(f"Energy Usage List (per second): {eval_tracker.energy_usage}")
    print(f"Total Time: {eval_tracker.total_time}s")
    print(f"Used Energy: {eval_tracker.used_energy} kWh")


if __name__ == "__main__":
    advanced_demo()



from abc import ABC, abstractmethod


class PowerMonitor(ABC):
    """
    The blueprint class for the platform specific subclasses.
    """

    @abstractmethod
    def get_current_power(self):
        """Returns the power usage in Watts."""
        pass

    @abstractmethod
    def get_gpu_utilization(self):
        """Returns GPU utilization as a percentage."""
        pass



import os

import torch
import torch.nn as nn
import torch.nn.functional as f
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms

from ecotorch import EvalTracker, TrainTracker, evaluate, train

# Deletable, I mostly run my codes in Terminal, so it is cleaner from me to run the tests this way
os.system("clear" if os.name != "nt" else "cls")


EPOCH = 40
BATCH_SIZE = 16
device = (
    "cuda"
    if torch.cuda.is_available()
    else ("mps" if torch.mps.is_available() else "cpu")
)


transform = transforms.ToTensor()

train_dataset = torchvision.datasets.MNIST(
    root="./data", train=True, download=True, transform=transform
)

test_dataset = torchvision.datasets.MNIST(
    root="./data", train=True, download=True, transform=transform
)

train_loader = torch.utils.data.DataLoader(
    train_dataset, batch_size=BATCH_SIZE, shuffle=True, num_workers=0
)

test_loader = torch.utils.data.DataLoader(
    test_dataset, batch_size=BATCH_SIZE, shuffle=False, num_workers=0
)


class TestNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.flatten = nn.Flatten()
        self.fc1 = nn.Linear(784, 32)
        self.fc2 = nn.Linear(32, 32)
        self.fc3 = nn.Linear(32, 10)

    def forward(self, x):
        x = self.flatten(x)
        x = f.relu(self.fc1(x))
        x = f.relu(self.fc2(x))
        x = self.fc3(x)
        return x


net = TestNet().to(device)
criterion = nn.CrossEntropyLoss().to(device)
optimizer = optim.SGD(net.parameters(), lr=0.001, momentum=0.9)

with TrainTracker(
    epochs=EPOCH, model=net, train_dataloader=train_loader
) as train_tracker:
    net, final_loss, first_loss, _, _ = train(
        net, criterion, optimizer, train_loader, EPOCH, device
    )
print(
    train_tracker.calculate_efficiency_score(
        initial_loss=first_loss, final_loss=final_loss
    )
)

print("Finished training")

with EvalTracker(
    test_dataloader=test_loader, train_tracker=train_tracker
) as eval_tracker:
    acc = evaluate(net, test_loader, device)

print(eval_tracker.calculate_efficiency_score(accuracy=acc))

print(f"Accuracy: {acc*100}%")



from pathlib import Path

import pandas as pd

from ._geolocator import convert_country_to_iso, get_location


class DataHandler:
    def __init__(self):
        base_path = Path(__file__).parent
        csv_path = base_path / "co2intensity.csv"
        self.df = pd.read_csv(csv_path)
        self._country = get_location()

    def get_intensity(self, country: str = None) -> float:
        if not self._country or country:
            if country is None:
                _country_code = get_location()
            else:
                _country_code = convert_country_to_iso(country)
            if _country_code is None:
                raise ValueError("Cannot find country code!")
        else:
            _country_code = self._country

        if _country_code == "World":
            _latest_year = self._get_latest_year_for_world
            mask = (
                (self.df["Area"] == _country_code)
                & (self.df["Variable"] == "CO2 intensity")
                & (self.df["Year"] == _latest_year)
            )
            data = self.df.loc[mask, "Value"]

            if data.empty:
                raise ValueError(
                    "Couldn't find values for the world, please provide a country!"
                )
            return float(data.iloc[0])

        _latest_year = self._get_latest_year_by_country_code(_country_code)
        if _latest_year is None:
            raise ValueError("No data for the country!")

        mask = (
            (self.df["ISO 3 code"] == _country_code)
            & (self.df["Variable"] == "CO2 intensity")
            & (self.df["Year"] == _latest_year)
        )
        data = self.df.loc[mask, "Value"]

        if data.empty:
            raise ValueError("Invalid country name!")
        return float(data.iloc[0])

    def _get_latest_year_by_country_code(self, country_code: str) -> int | None:
        _max_year = int(max(self.df["Year"].unique().tolist()))
        for i in range(_max_year, _max_year - 4, -1):
            _mask = (
                (self.df["ISO 3 code"] == country_code)
                & (self.df["Variable"] == "CO2 intensity")
                & (self.df["Year"] == i)
            )
            _data = self.df.loc[_mask, "Value"]
            if _data.empty:
                continue
            return i
        return None

    @property
    def _get_latest_year_for_world(self) -> int | None:
        _max_year = int(max(self.df["Year"].unique().tolist()))
        for i in range(_max_year, _max_year - 4, -1):
            _mask = (
                (self.df["Area"] == "World")
                & (self.df["Variable"] == "CO2 intensity")
                & (self.df["Year"] == i)
            )
            _data = self.df.loc[_mask, "Value"]
            if _data.empty:
                continue
            return i
        return None

    @property
    def countries(self) -> list:
        return sorted(self.df["Area"].unique().tolist())



import platform
import sys
from typing import TYPE_CHECKING, Type, Union

from .interface import PowerMonitor

if TYPE_CHECKING:
    from .apple import Monitor as AppleMonitor


def get_monitor_class() -> Union[Type[PowerMonitor], Type["AppleMonitor"]]:
    system = platform.system()

    if system == "Darwin":
        try:
            from .apple import Monitor as AppleMonitor

            return AppleMonitor
        except ImportError:
            # Fallback or error if the C++ build failed
            raise ImportError("EcoTorch C++ extension for macOS not found.")

    elif system == "Linux":
        from .linux import LinuxMonitor

        return LinuxMonitor

    elif system == "Windows":
        from .windows import WindowsMonitor

        return WindowsMonitor

    else:
        raise OSError(f"Unsupported operating system: {system}")


Monitor = get_monitor_class()



import json
from pathlib import Path


class Extractor:
    def __init__(self) -> None:
        base_path = Path(__file__).parent
        file_path = base_path / "electricity_price.json"
        self.prices = self._load_json(file_path)

    def _load_json(self, path) -> dict:
        try:
            with open(path, "r", encoding="utf-8") as file:
                data = json.load(file)
        except FileNotFoundError:
            raise FileNotFoundError("Internal error, please contact the developers.")
        except json.JSONDecodeError as e:
            raise json.JSONDecodeError(
                "Internal error, please contact developers.", e.doc, e.pos
            )
        return data

    def calculate_cost(self, country_code, kwh) -> float:
        return round(self._get_price(country_code) * kwh, 4)

    def _get_price(self, iso3) -> float:
        return self.prices[iso3]



import math
import threading
import time
from abc import ABC, abstractmethod
from enum import Enum, auto

import torch
import torch.nn as nn
import torch.optim as optim

from ._geolocator import get_location
from .datahandler import DataHandler
from .wallet import Extractor
from .watcher import Monitor


def _transfer_optimizer_to_device(optimizer: optim.Optimizer, device: str) -> None:
    """
    Moves an existing optimizer to a specific device.

    :param optimizer: The pytorch optimizer.
    :type optimizer: optim.Optimizer
    :param device: The device we move the optimizer to.
    :type device: str
    :return:
    """
    for param in optimizer.state.values():
        # move state to the device
        if isinstance(param, torch.Tensor):
            param.data = param.data.to(device)
            if param._grad is not None:
                param._grad.data = param._grad.data.to(device)
        elif isinstance(param, dict):
            for subparameter in param.values():
                if isinstance(subparameter, torch.Tensor):
                    subparameter.data = subparameter.data.to(device)
                    if subparameter._grad is not None:
                        subparameter._grad.data = subparameter._grad.data.to(device)


def evaluate(
    model: nn.Module, test_loader: torch.utils.data.DataLoader, device: str
) -> float:
    """
    Base implementation of an accuracy-based evaluation.
    :param model: The Pytorch model for the test
    :type model: nn.Module
    :param test_loader: The pytorch DataLoader for the test dataset
    :type test_loader: torch.utils.data.DataLoader
    :param device: The device the evaluation will run on
    :type device: str
    :return: The accuracy of the model on the test dataset
    :rtype: float
    """
    correct = 0
    total = 0
    with torch.no_grad():
        for data in test_loader:
            numbers, labels = data
            numbers, labels = numbers.to(device), labels.to(device)
            outputs = model(numbers)
            _, predicted = torch.max(outputs, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

    return round(correct / total, 4)


def train(
    model: nn.Module,
    criterion: nn.Module,
    optimizer: optim.Optimizer,
    train_loader: torch.utils.data.DataLoader,
    epoch: int,
    device: str,
) -> tuple[nn.Module, float, float, nn.Module, optim.Optimizer]:
    """
    Trains a neural network model using the provided dataset and optimizer.

    :param model: The neural network model to train.
    :type model: nn.Module
    :param criterion: The loss function used to compute the error
        during training.
    :type criterion: nn.Module
    :param optimizer: The optimizer used to apply computed gradients
        during backpropagation.
    :type optimizer: optim.Optimizer
    :param train_loader: The data loader that provides the training
        dataset in mini-batches.
    :type train_loader: torch.utils.data.DataLoader
    :param epoch: The number of epochs (full dataset iterations) to
        train the model for.
    :type epoch: int
    :param device: The device the training will run on
    :type device: str
    :return: A dictionary containing the trained model.
    :rtype: tuple[nn.Module, float, nn.Module, optim.Optimizer]
    """
    model = model.to(device)
    criterion = criterion.to(device)
    _transfer_optimizer_to_device(optimizer, device)

    running_loss = 0.0
    first_loss = 0.0
    for e in range(epoch):
        running_loss = 0.0
        for i, data in enumerate(train_loader):
            inputs, labels = data
            inputs, labels = inputs.to(device), labels.to(device)

            optimizer.zero_grad()

            outputs = model(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            running_loss += loss.item()
            if first_loss == 0.0:
                first_loss = loss.item()

            if i % 500 == 499:
                print(f"[{e + 1}, {i + 1:5d}] loss: {running_loss / 500:.3f}")
                running_loss = 0.0
    return model, running_loss / 500, first_loss, criterion, optimizer


class ModeError(Exception):
    pass


class Mode(Enum):
    TRAIN = auto()
    EVAL = auto()
    OTHER = auto()


class Tracker(ABC):
    """
    Tracks energy usage and cost during a specific time window.

    This class is designed to monitor and collect data related to energy usage
    within a defined duration, using GPU monitoring. It provides an interface
    to perform initialization and properly handle resource cleanup through
    context management.

    :ivar energy_usage: A list that stores energy usage data collected during
        monitoring.
    :type energy_usage: list
    :ivar cost: The calculated cost based on the energy usage data.
    :type cost: float
    :ivar _start_time: The timestamp capturing the start of the monitoring period.
    :type _start_time: float
    :ivar _end_time: The timestamp capturing the end of the monitoring period.
    :type _end_time: float
    """

    def __init__(
        self,
        mode: Mode,
        model: nn.Module,
        epochs: int = None,
        train_dataloader: torch.utils.data.DataLoader = None,
        test_dataloader: torch.utils.data.DataLoader = None,
        data_handler: DataHandler = None,
        country: str = None,
    ) -> None:
        self.energy_usage = []
        self.cost = 0
        self.total_time = 0

        self._gpu_monitor = None
        if data_handler is None:
            self._handler = DataHandler()
        else:
            self._handler = data_handler
        self._mode = mode
        self._is_closed = False
        self._stop_event = threading.Event()
        self._start_time = 0
        self._end_time = 0
        self._model = model
        if self._mode == Mode.TRAIN:
            self._dataloader = train_dataloader
            self._epochs = epochs
        elif self._mode == Mode.EVAL:
            self._dataloader = test_dataloader
            self._epochs = 0
        self._country = country
        self._extractor = Extractor()

    def __enter__(self):
        self._start_time = time.time()
        self._gpu_monitor = _Monitor("GPU", self.energy_usage, self._stop_event)
        self._gpu_monitor.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop_event.set()
        self._gpu_monitor.join()

        self._end_time = time.time()
        self.total_time = round(self._end_time - self._start_time, 4)
        self._is_closed = True

        print(f"The block emitted {self._co2_emission}g CO2")
        print(f"The block took {self.total_time} second to execute.")
        print(f"The block used {self.used_energy} kwh to execute.")
        print(f"The block used {self.block_cost} USD to execute.")
        return False

    @abstractmethod
    def calculate_efficiency_score(
        self,
        accuracy: float = None,
        initial_loss: float = None,
        final_loss: float = None,
    ) -> float:
        """
        Calculate the efficiency score of a model.

        :param initial_loss: The first loss in the training.
        :param final_loss: The final loss in the training.
        :param accuracy: Accuracy of the model.
        :return: The efficiency score.
        """
        pass

    @property
    @abstractmethod
    def _calculate_c_score(self) -> int:
        pass

    @property
    def _calculate_kwh(self) -> float:
        if len(self.energy_usage) == 0:
            return 0.0
        return sum(self.energy_usage) / 3600000

    @property
    def _co2_emission(self) -> float:
        if self._country is not None:
            return round(
                self._calculate_kwh * self._handler.get_intensity(self._country), 4
            )
        return round(self._calculate_kwh * self._handler.get_intensity(), 4)

    @property
    def _data_handler(self) -> DataHandler:
        return self._handler

    @property
    def used_energy(self) -> float:
        return round(self._calculate_kwh, 4)

    @property
    def total_cost(self) -> float:
        """
        Returns the total cost in USD (rounded to 4 decimal places).
        """
        return self.block_cost

    @property
    def block_cost(self) -> float:
        if self._country is None:
            self._country = get_location()
        return self._extractor.calculate_cost(self._country, self._calculate_kwh)


class TrainTracker(Tracker):
    def __init__(
        self,
        model: nn.Module,
        epochs: int,
        train_dataloader: torch.utils.data.DataLoader,
        country: str = None,
    ) -> None:
        self._model = model
        self._epochs = epochs
        self._train_dataloader = train_dataloader
        self._country = country
        super().__init__(
            Mode.TRAIN,
            model,
            epochs=epochs,
            train_dataloader=train_dataloader,
            country=country,
        )

    def calculate_efficiency_score(
        self,
        accuracy: float = None,
        initial_loss: float = None,
        final_loss: float = None,
    ) -> float:
        """
        Calculate the efficiency score of a model.

        :param initial_loss: The first loss in the training.
        :param final_loss: The final loss in the training.
        :param accuracy: Accuracy of the model.
        :return: The efficiency score.
        """

        if accuracy is not None and not (0.0 <= accuracy <= 1.0):
            raise ValueError("Accuracy must be between 0 and 1 (inclusive).")

        if not self._is_closed:
            raise RuntimeError(
                "The Tracker did not finish tracking; call this only after the 'with' block completes."
            )

        if initial_loss is None:
            raise ValueError("Initial loss cannot be None in train mode!")
        if final_loss is None:
            raise ValueError("Final loss cannot be None in train mode!")

        _c_score = self._calculate_c_score
        if initial_loss == 0.0:
            _progress = 0.0
        else:
            _progress = max(0.0, 1.0 - (final_loss / initial_loss))
        _emission = self._co2_emission
        _penalty = math.exp(-1.0 * (_emission / _c_score))
        score = _progress * _penalty

        return round(score, 4)

    @property
    def _calculate_c_score(self) -> int:
        P = sum(p.numel() for p in self._model.parameters())
        D = getattr(self._dataloader.dataset, "__len__", lambda: 0)() * self._epochs
        return 6 * P * D

    @property
    def country(self) -> str:
        return self._country

    @property
    def model(self) -> nn.Module:
        return self._model


class EvalTracker(Tracker):
    def __init__(
        self,
        test_dataloader: torch.utils.data.DataLoader,
        train_tracker: TrainTracker = None,
        model: nn.Module = None,
        country: str = None,
    ) -> None:
        if train_tracker:
            if country is not None:
                self._country = country
            else:
                self._country = train_tracker.country
            super().__init__(
                Mode.EVAL,
                train_tracker.model,
                test_dataloader=test_dataloader,
                data_handler=train_tracker._data_handler,
                country=self._country,
            )
        else:
            if model is None:
                raise ValueError("Please provide a model or TrainTracker!")

            super().__init__(
                Mode.EVAL, model, test_dataloader=test_dataloader, country=country
            )

    def calculate_efficiency_score(
        self,
        accuracy: float = None,
        initial_loss: float = None,
        final_loss: float = None,
    ) -> float:
        """
        Calculate the efficiency score of a model.

        :param initial_loss: The first loss in the training.
        :param final_loss: The final loss in the training.
        :param accuracy: Accuracy of the model.
        :return: The efficiency score.
        """

        if accuracy is not None and not (0.0 <= accuracy <= 1.0):
            raise ValueError("Accuracy must be between 0 and 1 (inclusive).")

        if not self._is_closed:
            raise RuntimeError(
                "The Tracker did not finish tracking; call this only after the 'with' block completes."
            )

        if accuracy is None:
            raise ValueError("Accuracy cannot be None in eval mode!")

        _c_score = self._calculate_c_score
        _emission = self._co2_emission
        _penalty = math.exp(-1.0 * (_emission / _c_score))
        score = accuracy * _penalty

        return round(score, 4)

    @property
    def _calculate_c_score(self) -> int:
        P = sum(p.numel() for p in self._model.parameters())
        D = getattr(self._dataloader.dataset, "__len__", lambda: 0)()
        return 2 * P * D

    @property
    def country(self) -> str:
        return self._country


class _Monitor(threading.Thread):
    def __init__(self, name, data_list: list, stop_event: threading.Event) -> None:
        super().__init__()
        self._name = name
        self._data_list = data_list
        self._stop_event = stop_event
        self._watcher = Monitor()

    def run(self) -> None:
        while not self._stop_event.is_set():
            self._data_list.append(self._watcher.get_current_power())
            self._stop_event.wait(1)

    @property
    def data_list(self) -> list:
        return self._data_list



import pynvml

from .interface import PowerMonitor


class WindowsMonitor(PowerMonitor):
    def __init__(self):
        try:
            pynvml.nvmlInit()
            self.device_count = pynvml.nvmlDeviceGetCount()
        except pynvml.NVMLError:
            self.device_count = 0

    def get_current_power(self) -> float:
        if self.device_count == 0:
            return 0.0

        total_power_watts = 0.0

        for i in range(self.device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            power_milliwatts = pynvml.nvmlDeviceGetPowerUsage(handle)

            total_power_watts += power_milliwatts / 1000.0

        return total_power_watts

    def get_gpu_utilization(self):
        if self.device_count == 0:
            return "0%"

        total_util = 0.0
        for i in range(self.device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
            total_util += utilization.gpu

        avg_util = total_util / self.device_count
        return f"{round(avg_util, 1)}%"
